/******************************************************************************/
/*                                                                            */
/*   TRAFFIC.C:  Traffic Light Controller using STC32                         */
/*                                                                            */
/******************************************************************************/

char code menu[] =
    "\n"
    "+***** TRAFFIC LIGHT CONTROLLER using STC32 FreeRTOS **********+\n"
    "| This program is a simple Traffic Light Controller.  Between  |\n"
    "| start time and end time the system controls a traffic light  |\n"
    "| with pedestrian self-service.  Outside of this time range    |\n"
    "| the yellow caution lamp is blinking.                         |\n"
    "+ command -+ syntax -----+ function ---------------------------+\n"
    "| Display  | D           | display times                       |\n"
    "| Quit     |ESC(0x1B)or Q| Quit display times                  |\n"
    "| Time     | T hh:mm:ss  | set clock time                      |\n"
    "| Start    | S hh:mm:ss  | set start time                      |\n"
    "| End      | E hh:mm:ss  | set end time                        |\n"
    "+----------+-------------+-------------------------------------+\n";
#include "FreeRTOS.h"
#include "task.h"
#include "semphr.h"                    //źͷļ
#include <stdio.h>                    /* standard I/O .h-file                 */
#include <ctype.h>                    /* character functions                  */
#include <string.h>                   /* string and memory functions          */

#include "traffic.h"                  /* project specific header file         */
#include "STC32G_GPIO.h"

struct time ctime = { 12,  0,  0 };   /* storage for clock time values        */
struct time start = {  7, 30,  0 };   /* storage for start time values        */
struct time end   = { 18, 30,  0 };   /* storage for end   time values        */

sbit  red    = P2 ^ 0;                /* I/O Pin:  red    lamp output         */
sbit  yellow = P2 ^ 1;                /* I/O Pin:  yellow lamp output         */
sbit  green  = P2 ^ 2;                /* I/O Pin:  green  lamp output         */
sbit  stop   = P2 ^ 3;                /* I/O Pin:  stop   lamp output         */
sbit  walk   = P2 ^ 4;                /* I/O Pin:  walk   lamp output         */
sbit  key    = P3 ^ 2;                /* I/O Pin:  self-service key input     */



#define red_on()  red = 0
#define red_off()  red = 1
#define yellow_on()  yellow = 0
#define yellow_off()  yellow = 1
#define green_on()  green = 0
#define green_off()  green = 1
#define stop_on()  stop = 0
#define stop_off()  stop = 1
#define walk_on()  walk = 0
#define walk_off()  walk = 1
#define key_pressed() key == 0


SemaphoreHandle_t xSemaphoreKEY;    //  ź
SemaphoreHandle_t xSemaphoreCOMMAND;//  COMMANDź

TaskHandle_t xHandleBlink = NULL;//blinking
TaskHandle_t xHandleLights = NULL;//lights

char xdata cmdline[16];                /* storage for command input line       */

void command ( void *pvParameters );
void lights ( void *pvParameters );
void key_task(void *pvParameters);
void clock ( void *pvParameters );

/******************************************************************************/
/*        Task 0 'init': Initialize                                           */
/******************************************************************************/
void traffic_init ( void )
{
    /* program execution starts here        */
	P3_MODE_IO_PU(GPIO_Pin_2);		    //P3 Ϊ׼˫
	P2_MODE_IO_PU(GPIO_Pin_All);		//P2 Ϊ׼˫
   
    xSemaphoreKEY = xSemaphoreCreateBinary();       //ź
    xSemaphoreCOMMAND = xSemaphoreCreateBinary();   //COMMANDź

    xTaskCreate((TaskFunction_t )command,
                (const char *    )"CMD",
                (uint16_t       )configDEFAULT_STACK_SIZE,
                (void *          )NULL,
                (UBaseType_t    )(configDEFAULT_PRIORITIES),
                (TaskHandle_t *  )NULL);
    xTaskCreate((TaskFunction_t )key_task,
                (const char *    )"key",
                (uint16_t       )configDEFAULT_STACK_SIZE,
                (void *          )NULL,
                (UBaseType_t    )(configDEFAULT_PRIORITIES),
                (TaskHandle_t *  )NULL);
    xTaskCreate((TaskFunction_t )lights,
                (const char *    )"lights",
                (uint16_t       )configDEFAULT_STACK_SIZE,
                (void *          )NULL,
                (UBaseType_t    )(configDEFAULT_PRIORITIES + 1),
                (TaskHandle_t *  )&xHandleLights);
    xTaskCreate((TaskFunction_t )clock,
                (const char *    )"clock",
                (uint16_t       )configDEFAULT_STACK_SIZE,
                (void *          )NULL,
                (UBaseType_t    )(configDEFAULT_PRIORITIES + 1),
                (TaskHandle_t *  )NULL);

}


bit display_time = 0;                 /* flag:  signal cmd state display_time */

/******************************************************************************/
/*        Task 2 'clock'                                                      */
/******************************************************************************/
void clock ( void *pvParameters )
{
    UNUSED(pvParameters);
    while (1)                           /* clock is an endless loop             */
    {
        if (++ctime.sec == 60)            /* calculate the second                 */
        {
            ctime.sec = 0;
            if (++ctime.min == 60)          /* calculate the minute                 */
            {
                ctime.min = 0;
                if (++ctime.hour == 24)       /* calculate the hour                   */
                {
                    ctime.hour = 0;
                }
            }
        }
        if (display_time)                 /* if command_status == display_time    */
        {
            xSemaphoreGive(xSemaphoreCOMMAND);//źʱѾ
        }
        vTaskDelay(1000);//ʵ1ʱ
    }
}


struct time rtime;                    /* temporary storage for entry time     */

/******************************************************************************/
/*        readtime: convert line input to time values & store in rtime        */
/******************************************************************************/
bit readtime (char xdata *buffer)
{
    unsigned char args;                          /* number of arguments         */

    rtime.sec = 0;                               /* preset second               */
    args = sscanf (buffer, "%bd:%bd:%bd",        /* scan input line for         */
                   &rtime.hour,                  /* hour, minute and second     */
                   &rtime.min,
                   &rtime.sec);

    if (rtime.hour > 23  ||  rtime.min > 59  ||  /* check for valid inputs      */
            rtime.sec > 59   ||  args < 2        ||  args == EOF)
    {
        printf ("\n*** ERROR: INVALID TIME FORMAT\n");
        return (0);
    }
    return (1);
}



#define ESC  0x1B                     /* ESCAPE character code                */

static bit   escape;                  /* flag: mark ESCAPE character entered  */

/******************************************************************************/
/*        Task 6 'get_escape': check if ESC (escape character) was entered    */
/******************************************************************************/
void get_escape ( void *pvParameters )
{
    UNUSED(pvParameters);
    while (1)                                    /* endless loop                */
    {
        char c = _getkey ();
        if ( c == ESC || c == 'Q')  escape = 1;        /* set flag if ESC entered     */
        if (escape)                                /* if escape flag send signal  */
        {
            xSemaphoreGive(xSemaphoreCOMMAND);              /* to task 'command'           */
        }
    }
}


/******************************************************************************/
/*        Task 1 'command': command processor */
/******************************************************************************/
void command ( void *pvParameters )
{
    unsigned char i;
    TaskHandle_t xHandleESC = NULL;
    UNUSED(pvParameters);
    vTaskDelay(1000);//ʱ1룬ΪغԶ򿪴ھܿʾ
    printf (menu);                               /* display command menu        */
    while (1)                                    /* endless loop                */
    {
        printf ("\nCommand: ");                    /* display prompt              */
        gets ((char xdata *)&cmdline, sizeof (cmdline));        /*ȡ룬עҪ\nβ*/
        for (i = 0; cmdline[i] != 0; i++)           /* תдĸ    */
        {
            cmdline[i] = toupper(cmdline[i]);
        }

        for (i = 0; cmdline[i] == ' '; i++);        /* ȥǰĿַ*/

        switch (cmdline[i])                         /*  */
        {
            case 'D':                                /* Display Time Command        */
                printf ("Start Time: %02d:%02d:%02d    "
                        "End Time: %02d:%02d:%02d\n",
                        start.hour, start.min, start.sec,
                        end.hour,   end.min,   end.sec);
                printf ("                        type ESC or Q to abort\n");

                /*  */
                xTaskCreate((TaskFunction_t )get_escape,
                            (const char *    )"ESC",
                            (uint16_t       )configDEFAULT_STACK_SIZE,
                            (void *          )NULL,
                            (UBaseType_t    )(configDEFAULT_PRIORITIES),
                            (TaskHandle_t *  )&xHandleESC);
                escape = 0;                            /* clear escape flag           */
                display_time = 1;                      /* set display time flag       */
                xSemaphoreTake(xSemaphoreCOMMAND, 0);            /* clear pending signals       */
                while (!escape)                        /* while no ESC entered        */
                {
                    printf ("Clock Time: %02d:%02d:%02d\n",      /* display time     */
                            ctime.hour, ctime.min, ctime.sec);
                    xSemaphoreTake(xSemaphoreCOMMAND, portMAX_DELAY );      //ȴʱӷź
                }
                vTaskDelete(xHandleESC);//ɾESC
                display_time = 0;                      /* clear display time flag     */
                printf ("\n\n");
                break;

            case 'T':                                /* Set Time Command            */
                if (readtime (&cmdline[i + 1]))         /* read time input and         */
                {
                    ctime.hour = rtime.hour;             /* store in 'ctime'            */
                    ctime.min  = rtime.min;
                    ctime.sec  = rtime.sec;
                }
                break;

            case 'E':                                /* Set End Time Command        */
                if (readtime (&cmdline[i + 1]))         /* read time input and         */
                {
                    end.hour = rtime.hour;               /* store in 'end'              */
                    end.min  = rtime.min;
                    end.sec  = rtime.sec;
                }
                break;

            case 'S':                                /* Set Start Time Command */
                if (readtime (&cmdline[i + 1]))         /* read time input and         */
                {
                    start.hour = rtime.hour;             /* store in 'start'            */
                    start.min  = rtime.min;
                    start.sec  = rtime.sec;
                }
                break;

            default:                                 /* Error Handling              */
                printf (menu);                         /* display command menu        */
                break;
        }
    }
}


/******************************************************************************/
/*        signalon: check if clock time is between start and end              */
/******************************************************************************/
static bit signalon (void)
{
    if (memcmp (&start, &end, sizeof (struct time)) < 0)
    {
        if (memcmp (&start, &ctime, sizeof (struct time)) < 0  &&
                memcmp (&ctime, &end,   sizeof (struct time)) < 0)  return (1);
    }

    else
    {
        if (memcmp (&end,   &ctime, sizeof (start)) > 0  &&
                memcmp (&ctime, &start, sizeof (start)) > 0)  return (1);
    }
    return (0);                                  /* signal off, blinking on     */
}


/******************************************************************************/
/*        Task 3 'blinking': runs if current time is outside start & end time */
/******************************************************************************/
void blinking ( void *pvParameters )          /* blink yellow light          */
{
    UNUSED(pvParameters);
    red_off();
    yellow_off();
    green_off();
    stop_off();
    walk_off();

    while (1)                                    /* endless loop                */
    {
        yellow_on();                                /* yellow light on             */
        vTaskDelay(1500);
        yellow_off();                                /* yellow light off            */

        vTaskDelay(1500);
        if (signalon ())                           /* if blinking time over       */
        {
            // ,lights
            xTaskCreate((TaskFunction_t )lights,
                        (const char *    )"lights",
                        (uint16_t       )configDEFAULT_STACK_SIZE,
                        (void *          )NULL,
                        (UBaseType_t    )(configDEFAULT_PRIORITIES),
                        (TaskHandle_t *  )&xHandleLights);
            vTaskDelete(xHandleBlink);//ɾblinking
        }
    }
}


/******************************************************************************/
/*      Task 4 'lights': executes if current time is between start & end time */
/******************************************************************************/
void lights ( void *pvParameters )
{
    UNUSED(pvParameters);
    /* traffic light operation     */
    red_on();
    yellow_off();
    green_off();
    stop_on();
    walk_off();
    while (1)
    {
        vTaskDelay(1500);                   ////ʱ1.5̵
        if (!signalon ())                          /* if traffic signal time over */
        {
            // ,blinking
            xTaskCreate((TaskFunction_t )blinking,
                        (const char *    )"blink",
                        (uint16_t       )configDEFAULT_STACK_SIZE,
                        (void *          )NULL,
                        (UBaseType_t    )(configDEFAULT_PRIORITIES),
                        (TaskHandle_t *  )&xHandleBlink);
            vTaskDelete(xHandleLights);//ɾlights

        }
        yellow_on();                        //Ƶ
        vTaskDelay(1500);                   //ʱ1.5Ƶ
        red_off();                          //
        yellow_off();                       //Ƶ
        green_on();                         //̵ͨ
        xSemaphoreTake(xSemaphoreKEY, 0);   //ȡһ£൱źֹ֮ǰ˰
        vTaskDelay(2000);                   //ʱ2̵
        xSemaphoreTake(xSemaphoreKEY, 5000); //ȴź5볬ʱûа̵ƹ7룬а£̵ֻ2
        yellow_on();                        //Ƶ
        green_off();                        //̵
        vTaskDelay(1500);                   //ʱ1.5Ƶ
        red_on();                           //
        yellow_off();                       //Ƶ
        vTaskDelay(1500);                   //ʱ1.5
        stop_off();                         //е
        walk_on();                          //еͨе
        vTaskDelay(5000);                   //ȴ5
        stop_on();                          //е
        walk_off();                         //еͨе
    }
}


/******************************************************************************/
/*        Task 5 'key_task': 񣬵ʱź   */
/******************************************************************************/
void key_task(void *pvParameters)
{
    BaseType_t err;
    UNUSED(pvParameters);
    while (1)                                    /* endless loop                */
    {
        if (key_pressed())                       //а£һź
        {
            err = xSemaphoreGive(xSemaphoreKEY);//ź
//            if (err != pdTRUE)//жźǷͳɹ
//            {
//                printf("źʧ!\r\n");
//            }
//            else
//            {
//                printf("źͳɹ!\r\n");
//            }
        }
        vTaskDelay(10);     /* ʱ10 ticks   */
    }
}
