#include "FreeRTOS.h"
#include "task.h"
/*-----------------------------------------------------------
 *ʱ䣺2022-04-23
 *ߣ  QQ2334526772
 *޸ĺ֣uart1ͷ淶
 *----------------------------------------------------------*/


/**
 * λС
 * ʹ BUFFER_MASK (BUFFER_SIZE-1) ΪЧ&
 * MASK = 64-3=63=B00111111
 * С 2 8163264128256
*/
#define TX_BUFFER_SIZE 128
#define RX_BUFFER_SIZE 128

#if (TX_BUFFER_SIZE & (TX_BUFFER_SIZE - 1)) != 0
#error "TX_BUFFER_SIZE must be a power of two"
#endif

#if (RX_BUFFER_SIZE & (RX_BUFFER_SIZE - 1)) != 0
#error "RX_BUFFER_SIZE must be a power of two"
#endif


#define TX_BUFFER_MASK (TX_BUFFER_SIZE-1)
#define RX_BUFFER_MASK (RX_BUFFER_SIZE-1)


struct tx_ring_buffer_t
{
    /** Buffer memory. */
    char buffer[TX_BUFFER_SIZE];
    /** Index of tail. */
    volatile  uint8_t tail_index;
    /** Index of head. */
    volatile  uint8_t head_index;
};

struct rx_ring_buffer_t
{
    /** Buffer memory. */
    char buffer[RX_BUFFER_SIZE];
    /** Index of tail. */
    volatile  uint8_t tail_index;
    /** Index of head. */
    volatile  uint8_t head_index;
};

struct  tx_ring_buffer_t  xdata uart_tx;
struct  rx_ring_buffer_t  xdata uart_rx;

static volatile bit uart1_sending = 0;  //ǿձ־λ

/**
 * void uart1_init(uint32_t baud)
 * ʷʱʱ2
 * 1ʹ2ͬĲʿʹͬһʱ2
*/
void uart1_init(uint32_t baud)
{
    uint16_t time_value;

    time_value = (uint16_t)((655360 - MAIN_Fosc * 10 / 4 / baud) / 10);
    SCON = 0x50;		//8λ,ɱ䲨
    AUXR |= 0x01;		//1ѡʱ2Ϊʷ
    AUXR |= 0x04;		//ʱʱ1Tģʽ

    T2L = time_value;		//öʱʼֵ
    T2H = time_value >> 8;		//öʱʼֵ
    AUXR |= 0x10;		//ʱ2ʼʱ
    TI = 0;
    RI = 0;
    REN = 1;
    //ļxdataСʱ㣬ûãҪֶ
    uart_tx.tail_index = 0;
    uart_tx.head_index = 0;
    uart_rx.tail_index = 0;
    uart_rx.head_index = 0;

    ES = 1;
}

/*
uart1_write1ʵַ1һ֡ͨSBUF,֮жϷ
*/
uint8_t uart1_write(uint8_t c)
{
//    ES = 0;
    //գҲڷ
    if(uart1_sending == 0)
    {
        SBUF = c;   //һֽ  ֮ж
        uart1_sending = 1;//־λλ
    }
    else
    {
        uint8_t nextHead =  (uart_tx.head_index + 1) & TX_BUFFER_MASK;
     
        //ͻˣȴ
        //ע⣺ҪϵͳûʹôڴӡԣҪһwhileĴע͵
        while(nextHead == uart_tx.tail_index)
        {
            vTaskDelay(1);//Ҳʹõȴźʽ򵥷ʽϵͳʱ
        }
        /* 뻺 */
        uart_tx.buffer[uart_tx.head_index] = c;
        uart_tx.head_index = nextHead;
    }
//    ES = 1;
    return 1;
}

uint8_t uart1_read(void)
{
    uint8_t c;

    if(uart_rx.head_index == uart_rx.tail_index)
    {
        /* ͷβһû */
        return 0;
    }

    c = uart_rx.buffer[uart_rx.tail_index];
    uart_rx.tail_index = ((uart_rx.tail_index + 1) & RX_BUFFER_MASK);

    return c;
}

uint8_t uart1_available(void)
{
    return (uart_rx.head_index - uart_rx.tail_index) & RX_BUFFER_MASK;

}

void uart1_write_bytes(const char *dat, uint8_t size)
{
    uint8_t i;
    for(i = 0; i < size; i++)
    {
        uart1_write(dat[i]);
    }
}

void uart1_print(uint8_t *p)
{
    for (; *p != 0;	p++)  uart1_write(*p);
}

char uart1_getkey(void)
{
    while  (uart_rx.head_index == uart_rx.tail_index)  {
        vTaskDelay(1);//Եȴһźʽ򵥷ʽϵͳʱ
    }
    return uart1_read();
}
extern void delay_ms(uint16_t ms);
void UART1_ISR_Handler (void) interrupt UART1_VECTOR
{
    uint8_t c;
    static uint8_t auto_flag = 0;

//    EA = 0;//ֹжǶ
    if(RI)
    {
        RI = 0;//Ҫ
        c = SBUF;
        if(c == 0x7F)
        {
            if(++auto_flag == 8)
            {
                //Զ 7F 7F 7F 7F 7F 7F 7F 7F
                auto_flag = 0;
                IAP_CONTR = 0x60;
                while(1);
//                delay_ms(3000);
            }
        }
        else
        {
            auto_flag = 0;
        }
        // յݷŵ
        /*  */
        if(((uart_rx.head_index - uart_rx.tail_index) & RX_BUFFER_MASK) == RX_BUFFER_MASK)
        {
            /* ԭ*/
            /* tail_index ++ */
            uart_rx.tail_index = ((uart_rx.tail_index + 1) & RX_BUFFER_MASK);
        }
        /* 뻺 */
        uart_rx.buffer[uart_rx.head_index] = c;
        uart_rx.head_index = ((uart_rx.head_index + 1) & RX_BUFFER_MASK);
    }
    if(TI)
    {
        TI = 0;//Ҫ
        if(uart_tx.head_index == uart_tx.tail_index)
        {
            uart1_sending = 0;//־λ
        }
        else
        {
            SBUF = uart_tx.buffer[uart_tx.tail_index];
            uart_tx.tail_index = ((uart_tx.tail_index + 1) & TX_BUFFER_MASK);
        }
    }
//    EA = 1;
}
